/*
 * (C) Copyright 2014 Nuxeo SA (http://nuxeo.com/) and contributors.
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the GNU Lesser General Public License
 * (LGPL) version 2.1 which accompanies this distribution, and is available at
 * http://www.gnu.org/licenses/lgpl-2.1.html
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * Contributors:
 *     Bogdan Stefanescu
 *     Antoine Taillefer
 */
package org.nuxeo.ecm.automation.server.jaxrs;

import java.io.IOException;
import java.util.List;

import javax.mail.MessagingException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.ws.rs.core.MediaType;
import javax.ws.rs.core.Response;

import org.nuxeo.ecm.automation.core.util.BlobList;
import org.nuxeo.ecm.automation.core.util.Paginable;
import org.nuxeo.ecm.automation.core.util.RecordSet;
import org.nuxeo.ecm.automation.jaxrs.DefaultJsonAdapter;
import org.nuxeo.ecm.automation.jaxrs.JsonAdapter;
import org.nuxeo.ecm.automation.jaxrs.io.documents.MultipartBlobs;
import org.nuxeo.ecm.core.api.Blob;
import org.nuxeo.ecm.core.api.CoreSession;
import org.nuxeo.ecm.core.api.DocumentModel;
import org.nuxeo.ecm.core.api.DocumentModelList;
import org.nuxeo.ecm.core.api.DocumentRef;
import org.nuxeo.ecm.webengine.jaxrs.session.SessionFactory;

/**
 * @author <a href="mailto:bs@nuxeo.com">Bogdan Stefanescu</a>
 * @author <a href="mailto:ataillefer@nuxeo.com">Antoine Taillefer</a>
 */
public class ResponseHelper {

    private ResponseHelper() {
    }

    public static Response notFound() {
        return Response.status(404).build();
    }

    public static Response emptyContent() {
        return Response.status(204).build();
    }

    public static Response notAllowed() {
        return Response.status(401).build();
    }

    public static Response blob(Blob blob) {
        return blob(blob, HttpServletResponse.SC_OK);
    }

    public static Response blobs(List<Blob> blobs) throws MessagingException, IOException {
        return blobs(blobs, HttpServletResponse.SC_OK);
    }

    public static Response blob(Blob blob, int httpStatus) {
        String type = blob.getMimeType();
        if (type == null || "???".equals(type)) {
            type = MediaType.APPLICATION_OCTET_STREAM;
        }
        return Response.status(httpStatus).entity(blob).type(type).header("Content-Disposition",
                "attachment; filename=" + blob.getFilename()).build();
    }

    public static Response blobs(List<Blob> blobs, int httpStatus) throws MessagingException, IOException {
        MultipartBlobs multipartBlobs = new MultipartBlobs(blobs);
        return Response.status(httpStatus).entity(multipartBlobs).type(
                new BoundaryMediaType(multipartBlobs.getContentType())).build();
    }

    /**
     * @since 5.7.2
     */
    public static Object getResponse(Object result, HttpServletRequest request) throws MessagingException, IOException {
        return getResponse(result, request, HttpServletResponse.SC_OK);
    }

    /**
     * Handle custom http status.
     *
     * @since 7.1
     */
    public static Object getResponse(Object result, HttpServletRequest request, int httpStatus) throws IOException,
            MessagingException {
        if (result == null || "true".equals(request.getHeader("X-NXVoidOperation"))) {
            return emptyContent();
        }
        if (result instanceof Blob) {
            return blob((Blob) result);
        } else if (result instanceof BlobList) {
            return blobs((BlobList) result);
        } else if (result instanceof DocumentRef) {
            CoreSession session = SessionFactory.getSession(request);
            return Response.status(httpStatus).entity(session.getDocument((DocumentRef) result)).build();
        } else if (result instanceof DocumentModel || result instanceof DocumentModelList
                || result instanceof JsonAdapter || result instanceof RecordSet || result instanceof Paginable<?>) {
            return Response.status(httpStatus).entity(result).build();
        } else { // try to adapt to JSON
            return Response.status(httpStatus).entity(new DefaultJsonAdapter(result)).build();
        }
    }

    /**
     * @since 7.1
     */
    public static class BoundaryMediaType extends MediaType {
        private final String ctype;

        BoundaryMediaType(String ctype) {
            super("multipart", "mixed");
            this.ctype = ctype;
        }

        @Override
        public String toString() {
            return ctype;
        }
    }
}
